import datetime
import json
from typing import List, Optional, Union

import attr
from ioprocmeta.oep.sections import (
    _OEPContext,
    _OEPContributor,
    _OEPLicense,
    _OEPRessource,
    _OEPReview,
    _OEPSource,
    _OEPSpatial,
    _OEPSubject,
    _OEPTemporal,
)
from ioprocmeta.base import MetaBase


@attr.define(repr=False)
class OEPMeta_1_5_0(MetaBase):
    """This class represents the OEP Metadata format in version 1.5.0.
    Each section of the meta data format is described by an attrs class.
    The class supports writing of the OEP Metadata format to disk and parsing the format into
    a class instance.
    """

    name: Optional[str] = attr.ib(None)
    title: Optional[str] = attr.ib(None)
    id: Optional[str] = attr.ib(None)
    description: Optional[str] = attr.ib(None)
    language: List[str] = attr.ib(factory=list)
    subject: List[_OEPSubject] = attr.ib(factory=list)
    keywords: List[str] = attr.ib(factory=list)
    publicationDate: Optional[Union[str, datetime.date]] = attr.ib(None)
    context: _OEPContext = attr.ib(factory=_OEPContext)
    spatial: _OEPSpatial = attr.ib(factory=_OEPSpatial)
    temporal: _OEPTemporal = attr.ib(factory=_OEPTemporal)
    sources: List[_OEPSource] = attr.ib(factory=list)
    licenses: List[_OEPLicense] = attr.ib(factory=list)
    contributors: List[_OEPContributor] = attr.ib(factory=list)
    resources: List[_OEPRessource] = attr.ib(factory=list)
    review: _OEPReview = attr.ib(factory=_OEPReview)
    at_id: Optional[Union[str, int]] = attr.ib(None)
    at_context: Optional[str] = attr.ib(None)
    metaMetadata: dict = attr.field(
        default={
            "metadataVersion": "OEP-1.5.0",
            "metadataLicense": {
                "name": "CC0-1.0",
                "title": "Creative Commons Zero v1.0 Universal",
                "path": "https://creativecommons.org/publicdomain/zero/1.0/",
            },
        }
    )

    _comment: dict = attr.field(
        default={
            "metadata": "Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/oemetadata)",
            "dates": "Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)",
            "units": "Use a space between numbers and units (100 m)",
            "languages": "Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
            "licenses": "License name must follow the SPDX License List (https://spdx.org/licenses/)",
            "review": "Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/blob/master/data-review/manual/review_manual.md)",
            "null": "If not applicable use: null",
            "todo": "If a value is not yet available, use: todo",
        }
    )

    def add_subject(self, name, path):
        s = _OEPSubject(name, path)
        self.subject.append(s)
        return s

    def add_source(self, title, description, path):
        s = _OEPSource(title, description, path)
        self.sources.append(s)
        return s

    def add_license(self, name, title, path, instruction, attribution):
        lic = _OEPLicense(name, title, path, instruction, attribution)
        self.licenses.append(lic)
        return lic

    def add_contributor(self, title, email, date, _object, comment):
        c = _OEPContributor(title, email, date, _object, comment)
        self.contributors.append(c)
        return c

    def add_resource(self, profile, name, path, _format, encoding):
        r = _OEPRessource(profile, name, path, _format, encoding)
        self.resources.append(r)
        return r

    def add_language(self, name):
        self.language.append(name)

    def add_keyword(self, name):
        self.keywords.append(name)

    @staticmethod
    def type():
        return "oep150"

    @classmethod
    def _post_as_dict_cleanup(cls, data):
        data["@id"] = data["at_id"]
        del data["at_id"]
        data["@context"] = data["at_context"]
        del data["at_context"]
        return data

    @classmethod
    def _post_from_dict_cleanup(cls, data):
        data["at_id"] = data["@id"]
        del data["@id"]
        data["at_context"] = data["@context"]
        del data["@context"]
        return data

    def __repr__(self):
        return f"{self.metaMetadata['metadataVersion']} metadata information:\n{self.as_json()}"

@attr.define(repr=False)
class OEPMeta_1_6_0(OEPMeta_1_5_0):
    """This class represents the OEP Metadata format in version 1.6.0.
    Each section of the meta data format is described by an attrs class.
    The class supports writing of the OEP Metadata format to disk and parsing the format into
    a class instance.
    """

    metaMetadata: dict = attr.field(
        default={
            "metadataVersion": "OEP-1.6.0",
            "metadataLicense": {
                "name": "CC0-1.0",
                "title": "Creative Commons Zero v1.0 Universal",
                "path": "https://creativecommons.org/publicdomain/zero/1.0/",
            },
        }
    )

    @staticmethod
    def type():
        return "oep160"



import datetime
import pathlib as pt
from typing import List, Optional, Tuple, Union

import attr


@attr.define
class _OEPSubject:
    name: Optional[str] = attr.ib(None)
    path: Optional[Union[str, pt.Path]] = attr.ib(None)


@attr.define
class _OEPContext:
    homepage: Optional[str] = attr.ib(None)
    documentation: Optional[str] = attr.ib(None)
    sourceCode: Optional[str] = attr.ib(None)
    contact: Optional[str] = attr.ib(None)
    grantNo: Optional[str] = attr.ib(None)
    fundingAgency: Optional[str] = attr.ib(None)
    fundingAgencyLogo: Optional[str] = attr.ib(None)
    publisherLogo: Optional[str] = attr.ib(None)


@attr.define
class _OEPSpatial:
    location: Optional[str] = attr.ib(None)
    extent: Tuple = attr.ib(factory=lambda: (None, None))
    resolution: Tuple = attr.ib(factory=lambda: (None, None))
    resolution: Tuple = attr.ib(factory=lambda: (None, None))


@attr.define
class _OEPTimeseries:
    start: Optional[Union[datetime.date, str]] = attr.ib(None)
    end: Optional[Union[datetime.date, str]] = attr.ib(None)
    resolution: Optional[str] = attr.ib(None)
    alignment: Optional[str] = attr.ib(None)
    aggregationType: Optional[str] = attr.ib(None)


@attr.define
class _OEPTemporal:
    referenceDate: Optional[Union[datetime.date, str]] = attr.ib(None)
    timeseries: List[_OEPTimeseries] = attr.ib(factory=list)

    def add_timeseries(self, start, end, resolution, alignment, aggregationType):
        ts = _OEPTimeseries(start, end, resolution, alignment, aggregationType)
        self.timeseries.append(ts)
        return ts


@attr.define
class _OEPLicense:
    name: Optional[str] = attr.ib(None)
    title: Optional[str] = attr.ib(None)
    path: Optional[Union[pt.Path, str]] = attr.ib(None)
    instruction: Optional[str] = attr.ib(None)
    attribution: Optional[str] = attr.ib(None)


@attr.define
class _OEPSource:
    title: Optional[str] = attr.ib(None)
    description: Optional[str] = attr.ib(None)
    path: Optional[Union[pt.Path, str]] = attr.ib(None)
    licenses: List[_OEPLicense] = attr.ib(factory=list)

    def add_license(self, name, title, path, instruction, attribution):
        lic = _OEPLicense(name, title, path, instruction, attribution)
        self.licenses.append(lic)
        return lic


@attr.define
class _OEPContributor:
    title: Optional[str] = attr.ib(None)
    email: Optional[str] = attr.ib(None)
    date: Optional[Union[datetime.date, str]] = attr.ib(None)
    object: Optional[str] = attr.ib(None)
    comment: Optional[str] = attr.ib(None)


@attr.define
class _OEPReview:
    path: Optional[Union[pt.Path, str]] = attr.ib(None)
    badge: Optional[str] = attr.ib(None)


@attr.define
class _OEPValueReference:
    value: Optional[str] = attr.ib(None)
    name: Optional[str] = attr.ib(None)
    path: Optional[Union[str, pt.Path]] = attr.ib(None)


@attr.define
class _OEPIsAbout:
    name: Optional[str] = attr.ib(None)
    path: Optional[str] = attr.ib(None)


@attr.define
class _OEPField:
    name: Optional[str] = attr.ib(None)
    description: Optional[str] = attr.ib(None)
    type: Optional[str] = attr.ib(None)
    unit: Optional[str] = attr.ib(None)
    isAbout: List[_OEPIsAbout] = attr.ib(factory=list)
    valueReference: List[_OEPValueReference] = attr.ib(_OEPValueReference)

    def add_is_about(self, name, path):
        ia = _OEPIsAbout(name, path)
        self.isAbout.append(ia)
        return ia

    def add_value_reference(self, value, name, path):
        v = _OEPValueReference(value, name, path)
        self.valueReference.append(v)
        return v


@attr.define
class _OEPReference:
    resource: Optional[str] = attr.ib(None)
    fields: List[str] = attr.ib(factory=list)


@attr.define
class _OEPForeignKey:
    fields: List[str] = attr.ib(factory=list)
    reference: _OEPReference = attr.ib(factory=_OEPReference)

    def add_field(self, name):
        self.fields.append(name)


@attr.define
class _OEPDialect:
    delimeter: Optional[str] = attr.ib(None)
    decimalSeparator: str = attr.ib(".")


@attr.define
class _OEPSchema:
    fields: List[_OEPField] = attr.ib(factory=list)
    primaryKey: List[str] = attr.ib(factory=list)
    foreignKeys: List[_OEPForeignKey] = attr.ib(factory=list)
    dialect: _OEPDialect = attr.ib(factory=_OEPDialect)

    def add_primary_key(self, name):
        self.primaryKey.append(name)

    def add_field(self, name, description, type, unit):
        f = _OEPField(name, description, type, unit)
        self.fields.append(f)
        return f

    def add_foreign_key(self):
        fk = _OEPForeignKey()
        self.foreignKeys.append(fk)
        return fk


@attr.define
class _OEPRessource:
    profile: Optional[str] = attr.ib(None)
    name: Optional[str] = attr.ib(None)
    path: Optional[str] = attr.ib(None)
    format: Optional[str] = attr.ib(None)
    encoding: Optional[str] = attr.ib(None)
    schema: _OEPSchema = attr.ib(factory=_OEPSchema)
    dialect: _OEPDialect = attr.ib(factory=_OEPDialect)



import datetime
import pathlib as pt
from typing import Optional, Union

import cattr


class StructuringConverter:
    def __init__(self, cls):
        self._converter = cattr.Converter()
        self.__class = cls

        self._converter.register_structure_hook(
            Optional[Union[int, str]], lambda v, t: t(v) if v is not None else None
        )
        self._converter.register_structure_hook(Optional[Union[datetime.date, str]], self.__str_hook_1)
        self._converter.register_structure_hook(Optional[Union[str, pt.Path]], self.__str_hook_2)

        self._converter.register_unstructure_hook(StructuringConverter, lambda v: None)
        self._converter.register_unstructure_hook(Optional[Union[datetime.date, str]], self.__ustr_hook_1)
        self._converter.register_unstructure_hook(Optional[Union[str, pt.Path]], self.__ustr_hook_2)

    def __ustr_hook_1(self, v):
        return v.isoformat() if isinstance(v, datetime.date) else v

    def __ustr_hook_2(self, v):
        return v.as_posix() if isinstance(v, pt.Path) else v

    def __str_hook_1(self, v, t):
        if v is None:
            return None
        if isinstance(v, datetime.date):
            return v
        if isinstance(v, str):
            try:
                ret = datetime.datetime.strptime(v, "%Y-%m-%d")
            except:
                try:
                    ret = datetime.datetime.strptime(v, "%Y-%m-%dT%H:%M%z")
                except:
                    raise IOError(f'Unknown date format "{v}"')
            return ret
        raise IOError(f"malformed date input '{v}'")

    def __str_hook_2(self, v, t):
        return None if v is None else pt.Path(v)

    def unstructure(self, instance):
        return self._converter.unstructure(instance)

    def structure(self, data):
        return self._converter.structure(data, self.__class)



import abc
import pathlib as pt
import json
from ioprocmeta.converter import StructuringConverter
from functools import wraps


def ensure_structure_converter_is_initialized(m):
    @wraps(m)
    def __decorator__(*args, **kwargs):
        cls = args[0]
        if cls._converter is None:
            cls._converter = StructuringConverter(cls)
        return m(*args, **kwargs)

    return __decorator__


class MetaBase(metaclass=abc.ABCMeta):
    _converter: StructuringConverter = None

    @abc.abstractmethod
    def type(self):
        pass

    @ensure_structure_converter_is_initialized
    def as_json(self):
        res = self.as_dict()
        return json.dumps(res, indent=4)

    @ensure_structure_converter_is_initialized
    def as_dict(self):
        data = self._converter.unstructure(self)
        self._post_as_dict_cleanup(data)
        return data

    # attention: the order of these decorators matters, as otherwise cls is not available!
    @classmethod
    @ensure_structure_converter_is_initialized
    def from_dict(cls, data):
        cls._post_from_dict_cleanup(data)
        return cls._converter.structure(data)

    @classmethod
    def _post_as_dict_cleanup(cls, data):
        pass

    @classmethod
    def _post_from_dict_cleanup(cls, data):
        pass

    @abc.abstractstaticmethod
    def type():
        pass

    @classmethod
    def read_json(cls, path: pt.Path):
        if not path.exists():
            raise IOError(f'Input file "{path.as_posix()}" does not exist.')

        with path.open("r") as ipf:
            data = json.load(ipf)
        return cls.from_dict(data)

    def write_json(self, path: pt.Path):
        with path.open("w") as opf:
            opf.write(self.as_json())



import pathlib as pt
import pytest

import ioprocmeta

from ioprocmeta.base import MetaBase


@pytest.mark.parametrize("profile_name", {"oep150"})
def test_profile_registry(profile_name):

    assert profile_name in ioprocmeta.available_standard_formats

    assert isinstance(ioprocmeta.available_standard_formats[profile_name](), MetaBase)


def test_standard_formats():
    for iprofile, iclass in ioprocmeta.available_standard_formats.items():
        assert iprofile == iclass.type()
        assert issubclass(iclass, MetaBase)



from ioprocmeta.oep import OEPMeta_1_5_0


def test_empty_oep():
    m = OEPMeta_1_5_0()

    assert m.name is None
    assert m.title is None
    assert m.id is None
    assert m.description is None
    assert isinstance(m.language, list)
    assert len(m.language) == 0
    assert isinstance(m.subject, list)
    assert len(m.subject) == 0
    assert isinstance(m.keywords, list)
    assert len(m.keywords) == 0
    assert m.publicationDate is None
    assert m.context.homepage is None
    assert m.context.documentation is None
    assert m.context.sourceCode is None
    assert m.context.contact is None
    assert m.context.grantNo is None
    assert m.context.fundingAgency is None
    assert m.context.fundingAgencyLogo is None
    assert m.context.publisherLogo is None
    assert m.spatial.location is None
    assert (
        isinstance(m.spatial.extent, tuple)
        and len(m.spatial.extent) == 2
        and m.spatial.extent[0] is None
        and m.spatial.extent[1] is None
    )
    assert (
        isinstance(m.spatial.resolution, tuple)
        and len(m.spatial.resolution) == 2
        and m.spatial.resolution[0] is None
        and m.spatial.resolution[1] is None
    )
    assert m.temporal.referenceDate is None
    assert isinstance(m.temporal.timeseries, list)
    assert len(m.temporal.timeseries) == 0
    assert isinstance(m.sources, list)
    assert len(m.sources) == 0
    assert isinstance(m.licenses, list)
    assert len(m.licenses) == 0
    assert isinstance(m.contributors, list)
    assert len(m.contributors) == 0
    assert isinstance(m.resources, list)
    assert len(m.resources) == 0
    assert m.review.path is None
    assert m.review.badge is None
    assert isinstance(m.metaMetadata, dict)
    assert m.metaMetadata["metadataVersion"] == "OEP-1.5.0"
    assert isinstance(m.metaMetadata["metadataLicense"], dict)
    assert m.metaMetadata["metadataLicense"]["name"] == "CC0-1.0"
    assert m.metaMetadata["metadataLicense"]["title"] == "Creative Commons Zero v1.0 Universal"
    assert m.metaMetadata["metadataLicense"]["path"] == "https://creativecommons.org/publicdomain/zero/1.0/"
    assert isinstance(m._comment, dict)
    assert (
        m._comment["metadata"]
        == "Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/oemetadata)"
    )
    assert (
        m._comment["dates"]
        == "Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss±hh)"
    )
    assert m._comment["units"] == "Use a space between numbers and units (100 m)"
    assert m._comment["languages"] == "Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)"
    assert m._comment["licenses"] == "License name must follow the SPDX License List (https://spdx.org/licenses/)"
    assert (
        m._comment["review"]
        == "Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/blob/master/data-review/manual/review_manual.md)"
    )
    assert m._comment["null"] == "If not applicable use: null"
    assert m._comment["todo"] == "If a value is not yet available, use: todo"
    assert m.at_id is None
    assert m.at_context is None


def test_oep_as_dict():
    m = OEPMeta_1_5_0()
    m = m.as_dict()
    assert isinstance(m, dict)
    assert m["name"] is None
    assert m["spatial"]["extent"][0] is None


def test_oep_as_json():
    m = OEPMeta_1_5_0()
    out = m.as_json()
    ref = """{
    "name": null,
    "title": null,
    "id": null,
    "description": null,
    "language": [],
    "subject": [],
    "keywords": [],
    "publicationDate": null,
    "context": {
        "homepage": null,
        "documentation": null,
        "sourceCode": null,
        "contact": null,
        "grantNo": null,
        "fundingAgency": null,
        "fundingAgencyLogo": null,
        "publisherLogo": null
    },
    "spatial": {
        "location": null,
        "extent": [
            null,
            null
        ],
        "resolution": [
            null,
            null
        ]
    },
    "temporal": {
        "referenceDate": null,
        "timeseries": []
    },
    "sources": [],
    "licenses": [],
    "contributors": [],
    "resources": [],
    "review": {
        "path": null,
        "badge": null
    },
    "metaMetadata": {
        "metadataVersion": "OEP-1.5.0",
        "metadataLicense": {
            "name": "CC0-1.0",
            "title": "Creative Commons Zero v1.0 Universal",
            "path": "https://creativecommons.org/publicdomain/zero/1.0/"
        }
    },
    "_comment": {
        "metadata": "Metadata documentation and explanation (https://github.com/OpenEnergyPlatform/oemetadata)",
        "dates": "Dates and time must follow the ISO8601 including time zone (YYYY-MM-DD or YYYY-MM-DDThh:mm:ss\\u00b1hh)",
        "units": "Use a space between numbers and units (100 m)",
        "languages": "Languages must follow the IETF (BCP47) format (en-GB, en-US, de-DE)",
        "licenses": "License name must follow the SPDX License List (https://spdx.org/licenses/)",
        "review": "Following the OEP Data Review (https://github.com/OpenEnergyPlatform/data-preprocessing/blob/master/data-review/manual/review_manual.md)",
        "null": "If not applicable use: null",
        "todo": "If a value is not yet available, use: todo"
    },
    "@id": null,
    "@context": null
}"""

    for iref_line, iout_line in zip(ref.split("\n"), out.split("\n")):
        assert iref_line == iout_line


def test_setting_values():
    m = OEPMeta_1_5_0()
    s = m.add_source("test", None, None)
    assert s == m.sources[0]
    assert len(m.sources) == 1
    assert m.sources[0].title == "test"

    s = m.add_subject("test", None)
    assert s == m.subject[0]
    assert len(m.subject) == 1
    assert m.subject[0].name == "test"

    s = m.add_license("test", None, None, None, None)
    assert s == m.licenses[0]
    assert len(m.licenses) == 1
    assert m.licenses[0].name == "test"

    s = m.add_contributor("test", None, None, None, None)
    assert s == m.contributors[0]
    assert len(m.contributors) == 1
    assert m.contributors[0].title == "test"

    s = m.add_resource("test", None, None, None, None)
    assert s == m.resources[0]
    assert len(m.resources) == 1
    assert m.resources[0].profile == "test"

    m.add_language("test")
    assert len(m.language) == 1
    assert m.language[0] == "test"

    m.add_keyword("test")
    assert len(m.keywords) == 1
    assert m.keywords[0] == "test"

    s = m.temporal.add_timeseries(None, None, None, None, "test")
    assert s == m.temporal.timeseries[0]
    assert len(m.temporal.timeseries) == 1
    assert m.temporal.timeseries[0].aggregationType == "test"

    s = m.sources[0].add_license("test", None, None, None, None)
    assert s == m.sources[0].licenses[0]
    assert len(m.sources[0].licenses) == 1
    assert m.sources[0].licenses[0].name == "test"

    assert m.resources[0].schema.dialect.delimeter is None
    assert m.resources[0].schema.dialect.decimalSeparator == "."

    m.resources[0].schema.add_primary_key("test")
    f = m.resources[0].schema.add_field("test", None, None, None)
    fk = m.resources[0].schema.add_foreign_key()
    assert m.resources[0].schema.primaryKey[0] == "test"

    assert f == m.resources[0].schema.fields[0]
    assert m.resources[0].schema.fields[0].name == "test"
    assert len(m.resources[0].schema.fields) == 1

    assert fk == m.resources[0].schema.foreignKeys[0]
    assert len(m.resources[0].schema.foreignKeys) == 1

    m.resources[0].schema.foreignKeys[0].add_field("test")
    assert len(m.resources[0].schema.foreignKeys[0].fields) == 1
    assert m.resources[0].schema.foreignKeys[0].fields[0] == "test"
